"""Classes and functions for interacting with and creating notes"""

from __future__ import annotations
from dataclasses import dataclass
import os
from pathlib import Path
import re
from subprocess import DEVNULL, Popen
import tempfile
from time import localtime, strftime
from typing import Any, Optional, TYPE_CHECKING

from click import Abort
import readchar
from rich.columns import Columns
from rich.markdown import Markdown
from rich.table import Table
from rich.text import Text

from apyanki import cards
from apyanki.config import cfg
from apyanki.console import console, consolePlain
from apyanki.fields import (
    check_if_generated_from_markdown,
    check_if_inconsistent_markdown,
    convert_field_to_text,
    convert_text_to_field,
    img_paths_from_field,
    img_paths_from_field_latex,
    prepare_field_for_cli,
    prepare_field_for_cli_raw,
    toggle_field_to_markdown,
)
from apyanki.utilities import cd, choose, editor

if TYPE_CHECKING:
    from apyanki.anki import Anki
    from anki.notes import Note as ANote


class Note:
    """A Note wrapper class"""

    def __init__(self, anki: Anki, note: ANote) -> None:
        self.a = anki
        self.n = note
        note_type = note.note_type()
        if note_type:
            self.model_name = note_type["name"]
        else:
            self.model_name = "__invalid-note__"
        self.field_names = list(self.n.keys())
        self.suspended = any(c.queue == -1 for c in self.n.cards())

    def __repr__(self) -> str:
        """Convert note to Markdown format"""
        lines = [
            "# Note",
            f"model: {self.model_name}",
            f"tags: {self.get_tag_string()}",
        ]

        if self.a.n_decks > 1:
            lines += [f"deck: {self.get_deck()}"]

        if not any(check_if_generated_from_markdown(f) for f in self.n.values()):
            lines += ["markdown: false"]

        lines += [""]

        for name, field in self.n.items():
            lines.append(f"## {name}")
            lines.append(convert_field_to_text(field))
            lines.append("")

        return "\n".join(lines)

    def pprint(self, print_raw: bool = False, list_cards: bool = False) -> None:
        """Print to screen"""
        # pylint: disable=import-outside-toplevel
        from anki import latex

        header = f"[green]# Note (nid: {self.n.id})[/green]"
        if self.suspended:
            header += " [red](suspended)[/red]"

        created = strftime("%F %H:%M", localtime(self.n.id / 1000))
        modified = strftime("%F %H:%M", localtime(self.n.mod))
        columned = [
            f"[yellow]model:[/yellow] {self.model_name} ({len(self.n.cards())} cards)",
            f"[yellow]tags:[/yellow] {self.get_tag_string()}",
            f"[yellow]created:[/yellow] {created}",
            f"[yellow]modified:[/yellow] {modified}",
        ]
        if self.a.n_decks > 1:
            columned += ["[yellow]deck:[/yellow] " + self.get_deck()]

        if not list_cards:
            flagged = [
                cards.get_flag(c, str(c.template()["name"]))
                for c in self.n.cards()
                if c.flags > 0
            ]
            if flagged:
                columned += [f"[yellow]flagged:[/yellow] {', '.join(flagged)}"]

        consolePlain.print(header)
        consolePlain.print(Columns(columned, width=37))

        if list_cards:
            self.print_cards()

        console.print()
        imgs: list[Path] = []
        for name, field in self.n.items():
            is_markdown = check_if_generated_from_markdown(field)
            if is_markdown:
                name += " [italic](markdown)[/italic]"

            console.print(f"[blue]## {name}[/blue]")
            if print_raw:
                console.print(Markdown(prepare_field_for_cli_raw(field)))
            else:
                text = prepare_field_for_cli(field, is_markdown)
                if is_markdown:
                    console.print(Markdown(text))
                else:
                    console.print(text)
            console.print("")

            # Render LaTeX if necessary and fill list of LaTeX images
            note_type = self.n.note_type()
            if note_type:
                latex.render_latex(field, note_type, self.a.col)
                imgs += img_paths_from_field_latex(field, note_type, self.a)

        if imgs:
            console.print("[blue]## LaTeX sources[/blue]")
            for line in imgs:
                console.print("- " + str(line))
            console.print("")

    def print_cards(self) -> None:
        """Print list of cards to screen"""
        table = Table(
            show_edge=False,
            padding=(0, 3, 0, 0),
            highlight=True,
            box=None,
            header_style=None,
        )
        table.add_column("Card name", header_style="yellow", no_wrap=True)
        table.add_column("Due", justify="right", header_style="white")
        table.add_column("Interval", justify="right", header_style="white")
        table.add_column("Reps", justify="right", header_style="white")
        table.add_column("Lapses", justify="right", header_style="white")
        table.add_column("Factor", justify="right", header_style="white")
        for card in sorted(self.n.cards(), key=lambda x: x.factor):
            table.add_row(
                "- " + str(card.template()["name"]) + cards.get_flag(card),
                cards.get_due_days(card, self.a.today),
                str(card.ivl),
                str(card.reps),
                str(card.lapses),
                str(card.factor / 10.0),
            )
        console.print(table)

    def show_images(self) -> None:
        """Show in the fields"""
        note_type = self.n.note_type()
        if not note_type:
            return

        images: list[Path] = []
        for html in self.n.values():
            images += img_paths_from_field_latex(html, note_type, self.a)
            images += img_paths_from_field(html)

        with cd(self.a.col.media.dir()):
            for file in images:
                view_cmd = cfg["img_viewers"].get(
                    file.suffix[1:], cfg["img_viewers_default"]
                )
                Popen(view_cmd + [file], stdout=DEVNULL, stderr=DEVNULL)

    def edit(self) -> None:
        """Edit tags and fields of current note"""
        with tempfile.NamedTemporaryFile(
            mode="w+", dir=os.getcwd(), prefix="edit_note_", suffix=".md"
        ) as tf:
            tf.write(str(self))
            tf.flush()

            retcode = editor(tf.name)
            if retcode != 0:
                console.print(f"[red]Editor return with exit code {retcode}![/red]")
                return

            notes = markdown_file_to_notes(tf.name)

        if not notes:
            console.print("[red]Something went wrong when editing note![/red]")
            return

        if len(notes) > 1:
            added_notes = self.a.add_notes_from_list(notes[1:])
            console.print(
                f"[green]Added {len(added_notes)} new notes while editing.[/green]"
            )
            console.wait_for_keypress()

        note = notes[0]

        new_tags = note.tags.split()
        if new_tags != self.n.tags:
            self.n.tags = new_tags

        if note.deck is not None and note.deck != self.get_deck():
            self.set_deck(note.deck)

        for i, text in enumerate(note.fields.values()):
            self.n.fields[i] = convert_text_to_field(text, use_markdown=note.markdown)

        self.a.col.update_note(self.n)
        self.a.modified = True
        if self.n.dupeOrEmpty():
            console.print("The updated note is now a dupe!")
            console.wait_for_keypress()

    def delete(self) -> None:
        """Delete the note"""
        self.a.delete_notes(self.n.id)

    def has_consistent_markdown(self) -> bool:
        """Check if markdown fields are consistent with html values"""
        return any(check_if_inconsistent_markdown(f) for f in self.n.values())

    def change_model(self) -> bool:
        """Change the note type"""
        console.clear()
        console.print("[red]Warning![/red]")
        console.print(
            "The note type is changed by creating a new note with the selected "
            "type and then deleting the old note. This means that the review "
            "progress is lost!"
        )
        if not console.confirm("\nContinue?"):
            return False

        models = sorted(self.a.model_names)  # type: ignore[has-type]
        while True:
            console.clear()
            console.print("Please choose new model:")
            for n, m in enumerate(models):
                console.print(f"  {n+1}: {m}")
            index: int = console.prompt_int(">>> ", prompt_suffix="") - 1
            try:
                new_model = models[index]
                self.a.set_model(new_model)
                model = self.a.get_model(new_model)
                if not model:
                    continue
            except IndexError:
                continue

            break

        fields: dict[str, str] = {}
        first_field: str = model["flds"][0]
        for field_name in model["flds"]:
            fields[field_name] = ""

        for old_field_name, old_field in self.n.items():
            fields[first_field] += f"### {old_field_name}\n{old_field}\n"

        note_data = NoteData(
            model["name"],
            " ".join(self.n.tags),
            fields,
            any(check_if_generated_from_markdown(f) for f in self.n.values()),
        )

        new_note = note_data.add_to_collection(self.a)
        new_note.edit()
        self.a.delete_notes(self.n.id)

        return True

    def toggle_marked(self) -> None:
        """Toggle marked tag for note"""
        if "marked" in self.n.tags:
            self.n.remove_tag("marked")
        else:
            self.n.add_tag("marked")
        self.n.flush()
        self.a.modified = True

    def toggle_suspend(self) -> None:
        """Toggle suspend for note"""
        cids = [c.id for c in self.n.cards()]

        if self.suspended:
            self.a.col.sched.unsuspendCards(cids)
        else:
            self.a.col.sched.suspendCards(cids)

        self.suspended = not self.suspended
        self.a.modified = True

    def toggle_markdown(self, index: int | None = None) -> None:
        """Toggle markdown on a field"""
        if index is None:
            field_name = choose(self.field_names, "Toggle markdown for field:")
            index = self.field_names.index(field_name)

        self.n.fields[index] = toggle_field_to_markdown(self.n.fields[index])
        self.n.flush()
        self.a.modified = True

    def clear_flags(self) -> None:
        """Clear flags for note"""
        for c in self.n.cards():
            if c.flags > 0:
                c.flags = 0
                c.flush()
                self.a.modified = True

    def reset_progress(self) -> None:
        """Reset progress for a card"""
        card_list = {c.template()["name"]: c for c in self.n.cards()}
        if len(card_list) <= 1:
            card_name = next(iter(card_list))
        else:
            card_name = choose(list(card_list.keys()), "Select card to reset:")

        card = card_list[card_name]
        console.print("\n[magenta]Resetting progress for card:")
        cards.print_question(card)
        cards.print_answer(card)
        cards.print_stats(card)
        if not console.confirm("[red bold]Are you sure?"):
            return

        self.a.col.sched.schedule_cards_as_new(
            [card.id], restore_position=True, reset_counts=True
        )
        self.a.modified = True
        console.print("[magenta]The progress was reset.")
        console.wait_for_keypress()

    def get_deck(self) -> str:
        """Return which deck the note belongs to"""
        return self.a.col.decks.name(self.n.cards()[0].did)

    def set_deck(self, deck: str) -> None:
        """Move note to deck"""
        newdid = self.a.col.decks.id(deck)
        cids = [c.id for c in self.n.cards()]

        if cids and newdid:
            self.a.col.set_deck(cids, newdid)
            self.a.modified = True

    def set_deck_interactive(self) -> None:
        """Move note to deck, interactive"""
        console.clear()
        console.print("[white]Available decks:")
        for d in self.a.col.decks.all_names_and_ids(include_filtered=False):
            console.print(f"* {d.name}")
        console.print("* OTHER -> create new deck")

        try:
            newdeck = console.prompt("[white]Specify target deck")
        except Abort:
            return

        self.set_deck(newdeck)

    def get_tag_string(self) -> str:
        """Get tag string"""
        return ", ".join(self.n.tags)

    def review(
        self,
        i: Optional[int] = None,
        number_of_notes: Optional[int] = None,
        remove_actions: Optional[list[str]] = None,
    ) -> bool:
        """Interactive review of the note

        This method is used by the review command.

        if the arguments "i" and "number_of_notes" are supplied, then they are
        displayed to show review progress.

        The "remove_actions" argument can be used to remove a default action
        from the action menu.
        """

        actions = {
            "c": "Continue",
            "e": "Edit",
            "a": "Add new",
            "d": "Delete",
            "m": "Toggle markdown",
            "*": "Toggle marked",
            "z": "Toggle suspend",
            "p": "Toggle pprint",
            "F": "Clear flags",
            "R": "Reset progress",
            "f": "Show images",
            "E": "Edit CSS",
            "D": "Change deck",
            "N": "Change model",
            "s": "Save and stop",
            "v": "Show cards",
            "x": "Save and stop",
        }

        if remove_actions:
            actions = {
                key: val for key, val in actions.items() if val not in remove_actions
            }

        note_number_string = ""
        if i is not None:
            if number_of_notes:
                note_number_string = f" {i+1} of {number_of_notes}"
            else:
                note_number_string = f" {i+1}"

        menu = Columns(
            [
                f"[blue]{key}[/blue]: {value}"
                for key, value in actions.items()
                if key != "x"
            ],
            padding=(0, 2),
            title=Text(
                f"Reviewing note{note_number_string}",
                justify="left",
                style="white",
            ),
        )

        print_raw_fields = False
        refresh = True
        show_cards = cfg["review_show_cards"]
        while True:
            if refresh:
                console.clear()
                console.print(menu)
                console.print("")
                self.pprint(print_raw_fields, list_cards=show_cards)

            refresh = True
            choice = readchar.readchar()
            action = actions.get(choice)

            if action == "Continue":
                return True

            if action == "Edit":
                self.edit()
                continue

            if action == "Add new":
                notes = self.a.add_notes_with_editor(
                    tags=self.get_tag_string(),
                    model_name=self.model_name,
                    template=self,
                )

                console.print(f"Added {len(notes)} notes")
                console.wait_for_keypress()
                continue

            if action == "Delete" and console.confirm(
                "Are you sure you want to delete the note?"
            ):
                self.delete()
                return True

            if action == "Toggle markdown":
                self.toggle_markdown()
                continue

            if action == "Toggle marked":
                self.toggle_marked()
                continue

            if action == "Toggle suspend":
                self.toggle_suspend()
                continue

            if action == "Toggle pprint":
                print_raw_fields = not print_raw_fields
                continue

            if action == "Clear flags":
                self.clear_flags()
                continue

            if action == "Reset progress":
                self.reset_progress()
                continue

            if action == "Show images":
                self.show_images()
                refresh = False
                continue

            if action == "Edit CSS":
                self.a.edit_model_css(self.model_name)
                continue

            if action == "Change deck":
                self.set_deck_interactive()
                continue

            if action == "Change model":
                if self.change_model():
                    return True
                continue

            if action == "Save and stop":
                console.print("Stopped")
                return False

            if action == "Show cards":
                show_cards = not show_cards
                continue


@dataclass
class NoteData:
    """Dataclass to contain data for a single note"""

    model: str
    tags: str
    fields: dict[str, str]
    markdown: bool = True
    deck: Optional[str] = None

    def add_to_collection(self, anki: Anki) -> Note:
        """Add note to collection

        Returns: The new note
        """
        model = anki.set_model(self.model)
        model_field_names: list[str] = [field["name"] for field in model["flds"]]
        if len(model_field_names) != len(self.fields):
            console.print(f"Error: Not enough fields for model {self.model}!")
            anki.modified = False
            raise Abort()

        field_names = [x.replace(" (markdown)", "") for x in self.fields.keys()]
        for x, y in zip(model_field_names, field_names):
            if x != y:
                console.print("Warning: Inconsistent field names " f"({x} != {y})")

        notetype = anki.col.models.current(for_deck=False)
        new_note = anki.col.new_note(notetype)

        note_type = new_note.note_type()
        if self.deck is not None and note_type is not None:
            note_type["did"] = anki.deck_name_to_id[self.deck]  # type: ignore[has-type]

        new_note.fields = [
            convert_text_to_field(f, use_markdown=self.markdown)
            for f in self.fields.values()
        ]

        for tag in self.tags.strip().split():
            new_note.add_tag(tag)

        if not new_note.dupeOrEmpty():
            anki.col.addNote(new_note)
            anki.modified = True
        else:
            field_name, field_value = list(self.fields.items())[0]
            console.print("[red]Dupe detected, new_note was not added!")
            console.print(f"First field: {field_name}")
            console.print(f"First value: {field_value}")

        return Note(anki, new_note)


def markdown_file_to_notes(filename: str) -> list[NoteData]:
    """Parse note data from a Markdown file"""
    try:
        notes = [
            NoteData(
                model=x["model"],
                tags=x["tags"],
                fields=x["fields"],
                markdown=x["markdown"],
                deck=x["deck"],
            )
            for x in _parse_markdown_file(filename)
        ]
    except KeyError as e:
        console.print(f"Error {e.__class__} when parsing {filename}!")
        console.print("This may typically be due to bad Markdown formatting.")
        raise Abort() from e

    return notes


def _parse_markdown_file(filename: str) -> list[dict[str, Any]]:
    """Parse the content of a Markdown file

    This must adhere to the specification of {add_from_file} from cli.py!
    """
    defaults: dict[str, Any] = {
        "model": "Basic",
        "markdown": True,
        "tags": "",
        "deck": None,
    }
    with open(filename, "r", encoding="utf8") as f:
        for line in f:
            match = re.match(r"#+\s*.*", line)
            if match:
                break

            match = re.match(r"(\w+): (.*)", line)
            if match:
                k, v = match.groups()
                k = k.lower()
                v = v.strip()
                if k in ("tag", "tags"):
                    defaults["tags"] = v.replace(",", "")
                elif k in ("markdown", "md"):
                    defaults["markdown"] = v in ("true", "yes")
                else:
                    defaults[k] = v

    notes: list[dict[str, Any]] = []
    current_note: dict[str, Any] = {}
    current_field: Optional[str] = None
    is_in_codeblock = False
    with open(filename, "r", encoding="utf8") as f:
        for line in f:
            if is_in_codeblock:
                if current_field is not None:
                    current_note["fields"][current_field] += line
                match = re.match(r"```\s*$", line)
                if match:
                    is_in_codeblock = False
                continue

            match = re.match(r"```\w*\s*$", line)
            if match:
                is_in_codeblock = True
                if current_field is not None:
                    current_note["fields"][current_field] += line
                continue

            if current_note and current_field is None:
                match = re.match(r"(\w+): (.*)", line)
                if match:
                    k, v = match.groups()
                    k = k.lower()
                    v = v.strip()
                    if k in ("tag", "tags"):
                        current_note["tags"] = v.replace(",", "")
                    elif k in ("markdown", "md"):
                        current_note["markdown"] = v in ("true", "yes")
                    else:
                        current_note[k] = v

            match = re.match(r"(#+)\s*(.*)", line)
            if not match:
                if current_field is not None:
                    current_note["fields"][current_field] += line
                continue

            level, title = match.groups()

            if len(level) == 1:
                if current_note and current_field is not None:
                    current_note["fields"][current_field] = current_note["fields"][
                        current_field
                    ].strip()
                    notes.append(current_note)

                current_note = {"title": title, "fields": {}, **defaults}
                current_field = None
                continue

            if len(level) == 2:
                if current_field is not None:
                    current_note["fields"][current_field] = current_note["fields"][
                        current_field
                    ].strip()

                if title in current_note["fields"]:
                    console.print(f"Error when parsing {filename}!")
                    raise Abort()

                current_field = title
                current_note["fields"][current_field] = ""

    # Add remaining note to list
    if current_note and current_field is not None:
        current_note["fields"][current_field] = current_note["fields"][
            current_field
        ].strip()
        notes.append(current_note)

    return notes



"""Simple utility functions."""

from contextlib import contextmanager, redirect_stdout
from io import TextIOWrapper
import os
from subprocess import call
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import Any, Generator, Optional, TypeVar

import readchar

from apyanki.console import console


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath: str) -> None:
        self.newPath = os.path.expanduser(newPath)
        self.savedPath = ""

    def __enter__(self) -> None:
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        os.chdir(self.savedPath)


def editor(filepath: str) -> int:
    """Use EDITOR to edit file at given path"""
    return call([os.environ.get("EDITOR", "vim"), filepath])


def edit_text(input_text: str, prefix: str = "") -> str:
    """Use EDITOR to edit text (from a temporary file)"""
    if prefix:
        prefix = prefix + "_"

    with NamedTemporaryFile(mode="w+", prefix=prefix, suffix=".md") as tf:
        tf.write(input_text)
        tf.flush()
        editor(tf.name)
        tf.seek(0)
        edited_message = tf.read().strip()

    return edited_message


chooseType = TypeVar("chooseType")


def choose(items: list[chooseType], text: str = "Choose from list:") -> chooseType:
    """Choose from list of items"""
    console.print(text)
    for i, element in enumerate(items):
        console.print(f"{i+1}: {element}")

    index = _read_number_between(1, len(items)) - 1
    return items[index]


@contextmanager
def suppress_stdout() -> Generator[TextIOWrapper, Any, Any]:
    """A context manager that redirects stdout to devnull"""
    with open(os.devnull, "w", encoding="utf8") as fnull:
        with redirect_stdout(fnull) as out:
            yield out


def _read_number_between(first: int, last: int) -> int:
    """Read number from user input between first and last (inclusive)"""
    console.print("> ", end="")
    while True:
        choice_str = ""
        choice_int = 0
        choice_digits = 0
        max_digits = len(str(last))

        while choice_digits < max_digits:
            if choice_digits > 0 and int(choice_str + "0") > last:
                break

            char = readchar.readchar()
            try:
                _ = int(char)
            except ValueError:
                continue

            next_int = int(choice_str + char)
            if next_int > 0:
                console.print(char, end="")
                choice_str += char
                choice_int = next_int
                choice_digits += 1

        if first <= choice_int <= last:
            console.print("")
            return choice_int

        console.print(f"\nPlease type number between {first} and {last}!\n> ", end="")



"""Global console instance for CLI output"""

from typing import Any, Optional

import click
import readchar
from rich.console import Console
from rich.prompt import Confirm, IntPrompt, Prompt


class ApyConsole(Console):
    """A slightly enriched console for apy"""

    def wait_for_keypress(self) -> None:
        """Wait for keypress to continue."""
        console.print(
            "[white]Press [italic]any key[/italic] to continue ... [/white]", end=""
        )
        readchar.readchar()

    def prompt(self, prompt: str, **kwargs: Any) -> str:
        """Prompt for string."""
        p = Prompt(console=self)
        return p.ask(prompt, **kwargs)

    def prompt_int(
        self, prompt: str, suffix: Optional[str] = None, **kwargs: Any
    ) -> int:
        """Prompt for integer."""
        result: int

        if suffix is not None:
            result = click.prompt(prompt, prompt_suffix=suffix, type=int)
        else:
            result = IntPrompt(console=self).ask(prompt, **kwargs)

        return result

    def confirm(self, prompt: str, **kwargs: Any) -> bool:
        """Prompt for confirmation"""
        return Confirm(console=self).ask(prompt, **kwargs)


console = ApyConsole()
consolePlain = ApyConsole(highlight=False)



"""Package for interfacing and manipulating Anki decks"""

from importlib.metadata import version
import os

__version__ = version("apyanki")


# Reduce rust verbosity, unless already explicitly increased. Anki by default
# sets it to debug
if "RUST_LOG" not in os.environ:
    os.environ["RUST_LOG"] = "warn,anki::media=info,anki::sync=info,anki::dbcheck=info"



"""Simple module to load configuration from file"""

import json
import os
from pathlib import Path
from typing import Any, Optional


def get_base_path() -> Optional[str]:
    """Get base path on current system"""
    # If base_path not defined: Look in environment variables
    if path_as_str := os.environ.get("APY_BASE"):
        return path_as_str

    if path_as_str := os.environ.get("ANKI_BASE"):
        return path_as_str

    # Otherwise look in usual paths:
    # https://docs.ankiweb.net/files.html#file-locations
    if xdg_data_home := os.environ.get("XDG_DATA_HOME"):
        if (path := Path(xdg_data_home) / "Anki2").exists():
            return str(path)

    if (path := Path.home() / ".local/share/Anki2").exists():
        return str(path)

    if (path := Path.home() / "Library/Application Support/Anki2").exists():
        return str(path)

    return None


# Parse configuration file (if it exists)
cfg: dict[str, Any]
cfg_path = os.environ.get("APY_CONFIG", "~/.config/apy/apy.json")
cfg_file = Path(cfg_path).expanduser()
if cfg_file.exists():
    with cfg_file.open(encoding="utf8") as f:
        cfg = json.load(f)
else:
    cfg = {}


CFG_DEFAULT_VALUES: dict[str, Any] = {
    "base_path": None,
    "img_viewers": {
        "svg": ["display", "-density", "300"],
    },
    "img_viewers_default": ["feh"],
    "markdown_models": ["Basic"],
    "presets": {},
    "profile_name": None,
    "query": "tag:marked OR -flag:0",
    "review_show_cards": False,
}

# Ensure that cfg has required keys
for required, default in CFG_DEFAULT_VALUES.items():
    if required not in cfg:
        cfg[required] = default


if cfg["base_path"] is None:
    cfg["base_path"] = get_base_path()


# Ensure base path is a proper absolute path
if cfg["base_path"]:
    cfg["base_path"] = os.path.abspath(os.path.expanduser(cfg["base_path"]))


# Ensure that default preset is defined
if "default" not in cfg["presets"]:
    cfg["presets"]["default"] = {
        "model": "Basic",
        "tags": [],
    }



"""An Anki collection wrapper class."""

from __future__ import annotations
import os
from pathlib import Path
import pickle
import sqlite3
import tempfile
import time
from types import TracebackType
from typing import Any, Generator, Optional, Sequence, TYPE_CHECKING, Type

from click import Abort
from rich.progress import Progress, SpinnerColumn, TextColumn

from apyanki import cards
from apyanki.config import cfg
from apyanki.console import console
from apyanki.note import Note, NoteData, markdown_file_to_notes
from apyanki.utilities import cd, choose, editor, suppress_stdout

if TYPE_CHECKING:
    from anki.notes import NoteId
    from anki.models import NotetypeDict
    from anki.cards import CardId


class Anki:
    """My Anki collection wrapper class."""

    def __init__(
        self,
        base_path: Optional[str] = None,
        collection_db_path: Optional[str] = None,
        profile_name: Optional[str] = None,
        **_kwargs: dict[str, Any],
    ):
        self.modified = False

        self._meta = None
        self._collection_db_path = ""
        self._profile_name = profile_name
        self._profile = None

        self._init_load_profile(base_path, collection_db_path)
        self._init_load_collection()
        self._init_load_config()

        with suppress_stdout():
            self.today: int = self.col.sched.today

        self.model_name_to_id: dict[str, int] = {
            m["name"]: m["id"] for m in self.col.models.all()
        }
        self.model_names = list(self.model_name_to_id.keys())

        self.deck_name_to_id = {d["name"]: d["id"] for d in self.col.decks.all()}
        self.deck_names = self.deck_name_to_id.keys()
        self.n_decks: int = len(self.deck_names)

    def _init_load_profile(
        self, base_path_str: Optional[str], collection_db_path: Optional[str]
    ) -> None:
        """Load the Anki profile from database"""
        if base_path_str is None:
            if collection_db_path:
                self._collection_db_path = str(Path(collection_db_path).absolute())
                return

            console.print("Base path is not properly set!")
            raise Abort()

        base_path = Path(base_path_str)
        db_path = base_path / "prefs21.db"

        if not db_path.exists():
            console.print("Invalid base path!")
            console.print(f"path = {base_path.absolute()}")
            raise Abort()

        # Load metadata and profiles from database
        conn = sqlite3.connect(db_path)
        try:
            res = conn.execute(
                "select cast(data as blob) from profiles where name = '_global'"
            )
            self._meta = pickle.loads(res.fetchone()[0])

            profiles = conn.execute(
                "select name, cast(data as blob) from profiles where name != '_global'"
            ).fetchall()
        finally:
            conn.close()

        profiles_dict = {name: pickle.loads(data) for name, data in profiles}

        if self._profile_name is None:
            self._profile_name = self._meta.get(
                "last_loaded_profile_name", profiles[0][0]
            )

        self._collection_db_path = str(
            base_path / self._profile_name / "collection.anki2"
        )
        self._profile = profiles_dict[self._profile_name]

    def _init_load_collection(self) -> None:
        """Load the Anki collection"""
        # pylint: disable=import-outside-toplevel
        from anki.collection import Collection
        from anki.errors import DBError

        # Save CWD (because Anki changes it)
        save_cwd = os.getcwd()

        try:
            self.col = Collection(self._collection_db_path)
        except AssertionError as error:
            console.print("Path to database is not valid!")
            console.print(f"path = {self._collection_db_path}")
            raise Abort() from error
        except DBError as error:
            console.print("Database is NA/locked!")
            raise Abort() from error

        # Restore CWD (because Anki changes it)
        os.chdir(save_cwd)

    @staticmethod
    def _init_load_config() -> None:
        """Load custom configuration"""
        # pylint: disable=import-outside-toplevel
        from anki import latex

        # Update LaTeX commands
        # * Idea based on Anki addon #1546037973 ("Edit LaTeX build process")
        if "pngCommands" in cfg:
            latex.pngCommands = cfg["pngCommands"]
        if "svgCommands" in cfg:
            latex.svgCommands = cfg["svgCommands"]

    def __enter__(self) -> Anki:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if self.modified:
            console.print("Database was modified.")
            if self._profile is not None and self._profile["syncKey"]:
                console.print("[blue]Remember to sync!")

        self.col.close()

    def sync(self) -> None:
        """Sync collection to AnkiWeb"""
        # pylint: disable=import-outside-toplevel
        from anki.sync import SyncAuth

        if self._profile is None:
            return

        hkey = self._profile.get("syncKey")
        if not hkey:
            return

        auth = SyncAuth(
            hkey=hkey,
            endpoint=self._profile.get("currentSyncUrl")
            or self._profile.get("customSyncUrl")
            or None,
            io_timeout_secs=self._profile.get("networkTimeout") or 30,
        )

        if auth is None:
            return

        with Progress(
            TextColumn(
                "Syncing {task.fields[name]} [green]…[/green] {task.description}"
            ),
            SpinnerColumn(spinner_name="point", finished_text=""),
            console=console,
        ) as progress:
            t1 = progress.add_task("", total=None, name="deck")
            t2 = progress.add_task("", total=None, name="media")

            # Perform main sync
            with suppress_stdout():
                self.col.sync_collection(auth, True)
            progress.update(t1, total=1, completed=1, description="[green]done!")

            # Perform media sync
            with cd(self.col.media.dir()):
                status_str = ""
                self.col.sync_media(auth)
                try:
                    while True:
                        time.sleep(0.01)
                        status = self.col.media_sync_status()
                        if p := status.progress:
                            status_str = f"{p.added}, {p.removed}, {p.checked}".lower()
                            progress.update(t2, description=f"[blue]({status_str})")
                        if not status.active:
                            break

                except Exception as error:
                    if "sync cancelled" in str(error):
                        progress.update(
                            t2,
                            total=1,
                            completed=1,
                            description="[yellow]cancelled!",
                        )
                        return
                    raise Abort() from error

                progress.update(
                    t2,
                    total=1,
                    completed=1,
                    description=f"[blue]({status_str}) [green]done!",
                )

    def check_media(self) -> None:
        """Check media (will rebuild missing LaTeX files)"""
        # pylint: disable=import-outside-toplevel
        from anki.notes import NoteId

        with cd(self.col.media.dir()):
            with Progress(
                TextColumn("{task.description}"),
                SpinnerColumn(spinner_name="point", finished_text=""),
                console=console,
            ) as progress:
                t1 = progress.add_task("Checking media DB [green]… ", total=None)
                output = self.col.media.check()
                progress.update(
                    t1,
                    total=1,
                    completed=1,
                    description="Checking media DB [green]… done!",
                )

            if len(output.missing) + len(output.unused) == 0:
                console.print("[white]No unused or missing files found.")
                return

            for file in output.missing:
                console.print(f"[red]Missing: {file}")

            if len(output.missing) > 0 and console.confirm("Render missing LaTeX?"):
                out = self.col.media.render_all_latex()
                if out is not None:
                    nid = NoteId(out[0])
                    console.print(f"[red]Error processing note: {nid}")

                    if console.confirm("Review note?"):
                        note = Note(self, self.col.get_note(nid))
                        note.review()

            for file in output.unused:
                console.print(f"[red]Unused: {file}")

            if len(output.unused) > 0 and console.confirm("Delete unused media?"):
                for file in output.unused:
                    if os.path.isfile(file):
                        os.remove(file)

    def find_cards(self, query: str) -> Sequence[CardId]:
        """Find card ids in Collection that match query"""
        return self.col.find_cards(query)

    def find_notes(self, query: str) -> Generator[Note, None, None]:
        """Find notes in Collection and return Note objects"""
        return (
            Note(self, self.col.get_note(i)) for i in set(self.col.find_notes(query))
        )

    def delete_notes(self, ids: NoteId | list[NoteId]) -> None:
        """Delete notes by note ids"""
        if not isinstance(ids, list):
            ids = [ids]

        self.col.remove_notes(ids)
        self.modified = True

    def get_model(self, model_name: str) -> Optional[NotetypeDict]:
        """Get model from model name"""
        # pylint: disable=import-outside-toplevel
        from anki.models import NotetypeId

        model_id = self.model_name_to_id.get(model_name)
        if not isinstance(model_id, int):
            return None

        return self.col.models.get(NotetypeId(model_id))

    def set_model(self, model_name: str) -> NotetypeDict:
        """Set current model based on model name"""
        current = self.col.models.current(for_deck=False)
        if current["name"] == model_name:
            return current

        model = self.get_model(model_name)
        if model is None:
            console.print(f'Model "{model_name}" was not recognized!')
            raise Abort()

        self.col.models.set_current(model)
        return model

    def rename_model(self, old_model_name: str, new_model_name: str) -> None:
        """Rename a model"""
        model = self.get_model(old_model_name)
        if not model:
            console.print("Can't rename model!")
            console.print(f"No such model: {old_model_name}")
            raise Abort()

        # Change the name
        model["name"] = new_model_name

        # Update local storage
        self.model_name_to_id = {m["name"]: m["id"] for m in self.col.models.all()}
        self.model_names = list(self.model_name_to_id.keys())

        # Save changes
        self.col.models.update_dict(model)
        self.modified = True

    def list_tags(self) -> None:
        """List all tags"""
        tags = [(t, len(self.col.find_notes(f"tag:{t}"))) for t in self.col.tags.all()]
        width = len(max(tags, key=lambda x: len(x[0]))[0]) + 2
        filler = " " * (cfg["width"] - 2 * width - 8)

        for (t1, n1), (t2, n2) in zip(
            sorted(tags, key=lambda x: x[0]), sorted(tags, key=lambda x: x[1])
        ):
            console.print(f"{t1:{width}s}{n1:4d}{filler}{t2:{width}s}{n2:4d}")

    def change_tags(self, query: str, tags: str, add: bool = True) -> None:
        """Add/Remove tags from notes that match query"""
        note_ids = self.col.find_notes(query)
        if add:
            self.col.tags.bulk_add(note_ids, tags)
        else:
            self.col.tags.bulk_remove(note_ids, tags)

        self.modified = True

    def edit_model_css(self, model_name: str) -> None:
        """Edit the CSS part of a given model."""
        model = self.get_model(model_name)
        if not model:
            console.print(f"Could not find model: {model_name}!")
            return

        with tempfile.NamedTemporaryFile(
            mode="w+", prefix="_apy_edit_", suffix=".css", delete=False
        ) as tf:
            tf.write(model["css"])
            tf.flush()

            retcode = editor(tf.name)
            if retcode != 0:
                console.print(f"Editor return with exit code {retcode}!")
                return

            with open(tf.name, "r", encoding="utf8") as f:
                new_content = f.read()

        if model["css"] != new_content:
            model["css"] = new_content
            self.col.models.save(model, templates=True)
            self.modified = True

    def list_notes(self, query: str) -> None:
        """List notes that match a query"""
        for note in self.find_notes(query):
            cards.print_question(note.n.cards()[0])

    def list_cards(self, query: str, verbose: bool = False) -> None:
        """List cards that match a query"""
        for cid in self.find_cards(query):
            card = self.col.get_card(cid)
            cards.print_question(card)

            if verbose:
                cards.print_answer(card)
                cards.print_stats(card)

    def add_notes_with_editor(
        self,
        tags: str = "",
        model_name: Optional[str] = None,
        deck_name: Optional[str] = None,
        template: Optional[Note] = None,
    ) -> list[Note]:
        """Add new notes to collection with editor"""
        if template:
            input_string = str(template)
        else:
            if model_name is None or model_name.lower() == "ask":
                model_name = choose(sorted(self.model_names), "Choose model:")

            if deck_name is None:
                deck_name = self.col.decks.current()["name"]
            elif deck_name.lower() == "ask":
                deck_name = choose(sorted(self.deck_names), "Choose deck:")

            input_strings = [f"model: {model_name}"]

            if self.n_decks > 1:
                input_strings += [f"deck: {deck_name}"]

            input_strings += [f"tags: {tags}"]

            if model_name not in cfg["markdown_models"]:
                input_strings += ["markdown: false"]

            input_strings += ["\n# Note\n"]

            model = self.set_model(model_name)
            input_strings += [
                x
                for y in [[f'## {field["name"]}', ""] for field in model["flds"]]
                for x in y
            ]

            input_string = "\n".join(input_strings) + "\n"

        with tempfile.NamedTemporaryFile(
            mode="w+", prefix="apy_note_", suffix=".md", delete=False
        ) as tf:
            tf.write(input_string)
            tf.flush()
            retcode = editor(tf.name)

            if retcode != 0:
                console.print(f"Editor return with exit code {retcode}!")
                return []

            return self.add_notes_from_file(tf.name)

    def add_notes_from_file(
        self, filename: str, tags: str = "", deck: Optional[str] = None
    ) -> list[Note]:
        """Add new notes to collection from Markdown file"""
        notes = markdown_file_to_notes(filename)
        return self.add_notes_from_list(notes, tags, deck)

    def add_notes_from_list(
        self,
        parsed_notes: list[NoteData],
        tags: str = "",
        deck: Optional[str] = None,
    ) -> list[Note]:
        """Add new notes to collection from note list (from parsed file)"""
        notes = []
        for note in parsed_notes:
            if note.deck is None:
                note.deck = deck
            note.tags = f"{tags} {note.tags}"
            notes.append(note.add_to_collection(self))

        return notes

    def add_notes_single(
        self,
        field_values: list[str],
        markdown: bool,
        tags: str = "",
        model_name_in: Optional[str] = None,
        deck: Optional[str] = None,
    ) -> Note:
        """Add new note to collection from args"""
        model_name: str
        if model_name_in:
            model = self.set_model(model_name_in)
            model_name = model_name_in
        else:
            model = self.col.models.current(for_deck=False)
            model_name = model["name"]

        field_names: list[str] = [field["name"] for field in model["flds"]]
        fields = dict(zip(field_names, field_values))

        new_note = NoteData(model_name, tags, fields, markdown, deck)
        return new_note.add_to_collection(self)



"""Utility functions for working with Anki cards"""

from __future__ import annotations
from typing import TYPE_CHECKING

from rich.text import Text

from apyanki.console import console
from apyanki.fields import prepare_field_for_cli_oneline

if TYPE_CHECKING:
    from anki.cards import Card


def print_question(card: Card) -> None:
    """Print the card question"""
    question = Text("Q: ")
    question.stylize("yellow", 0, 2)
    question.append_text(
        Text.from_markup(prepare_field_for_cli_oneline(card.question()))
    )
    console.print(question.fit(console.width))


def print_answer(card: Card) -> None:
    """Print the card answer"""
    answer = Text("A: ")
    answer.stylize("yellow", 0, 2)
    answer.append_text(Text.from_markup(prepare_field_for_cli_oneline(card.answer())))
    console.print(answer.fit(console.width))


def print_stats(card: Card) -> None:
    """Print the card statistics"""
    cardtype = int(card.type)
    card_type = ["new", "learning", "review", "relearning"][cardtype]

    style = "green"
    console.print(
        Text.assemble(("model: ", style), card.note_type()["name"]),
        Text.assemble(("due: ", style), str(card.due)),
        Text.assemble(("type: ", style), card_type),
        Text.assemble(("ease: ", style), str(card.factor / 10)),
        Text.assemble(("lapses: ", style), str(card.lapses)),
        "\n",
    )


def get_flag(card: Card, text: str = "  ") -> str:
    """Get rich formatted flag of card"""
    style = {
        1: "red",
        2: "orange",
        3: "green",
        4: "blue",
        5: "pink1",
        6: "medium_turquoise",
        7: "purple",
    }.get(card.flags)

    if style:
        return f"[{style}]{text}[/{style}]"

    return ""


def get_due_days(card: Card, today: int) -> str:
    """Get number of days until card is due"""
    if card.type < 2:
        return "0"

    if card.type == 2:
        return str(card.due - today)

    return "?"
